@@ -56,12 +56,10 @@ from .util import (
     prompt_for_touch,
     EnumChoice,
     is_yk4_fips,
+    log_or_echo,
 )
-from .. import __version__
 from ..scancodes import encode, KEYBOARD_LAYOUT
 from ..otp import (
-    _PrepareUploadFailed,
-    _prepare_upload_key,
     is_in_fips_mode,
     generate_static_pw,
     parse_oath_key,
@@ -76,7 +74,6 @@ import logging
 import os
 import struct
 import click
-import webbrowser
 
 
 logger = logging.getLogger(__name__)
@@ -200,14 +197,27 @@ def otp(ctx, access_code):
 
 def _get_session(ctx, types=[OtpConnection, SmartCardConnection]):
     dev = ctx.obj["device"]
+
+    resolve_scp = ctx.obj.get("scp")
+    if resolve_scp:
+        if SmartCardConnection in types:
+            types = [SmartCardConnection]
+        else:
+            raise CliFail("SCP can only be used with SmartCardConnection")
+
     for conn_type in types:
         if dev.supports_connection(conn_type):
             conn = dev.open_connection(conn_type)
             ctx.call_on_close(conn.close)
-            return YubiOtpSession(conn)
+            if resolve_scp:
+                scp_params = resolve_scp(conn)
+            else:
+                scp_params = None
+            return YubiOtpSession(conn, scp_params)
+
     raise CliFail(
         "The connection type required for this command is not supported/enabled on the "
-        "YubiKey"
+        "YubiKey."
     )
 
 
@@ -243,9 +253,9 @@ def swap(ctx, force):
         err=True,
     )
 
-    click.echo("Swapping slots...")
     try:
         session.swap_slots()
+        click.echo("Slots swapped.")
     except CommandError:
         raise CliFail(_WRITE_FAIL_MSG)
 
@@ -282,6 +292,7 @@ def ndef(ctx, slot, prefix, ndef_type):
 
     try:
         session.set_ndef_configuration(slot, prefix, ctx.obj["access_code"], ndef_type)
+        click.echo("NDEF configuration updated.")
     except CommandError:
         raise CliFail(_WRITE_FAIL_MSG)
 
@@ -303,9 +314,9 @@ def delete(ctx, slot, force):
         abort=True,
         err=True,
     )
-    click.echo(f"Deleting the configuration in slot {slot}...")
     try:
         session.delete_slot(slot, ctx.obj["access_code"])
+        click.echo(f"Configuration slot {slot} deleted.")
     except CommandError:
         raise CliFail(_WRITE_FAIL_MSG)
 
@@ -366,7 +377,7 @@ def delete(ctx, slot, force):
     "--upload",
     is_flag=True,
     required=False,
-    help="upload credential to YubiCloud (opens a browser, can't be used with --force)",
+    hidden=True,
 )
 @click.option(
     "-O",
@@ -395,19 +406,23 @@ def yubiotp(
     Program a Yubico OTP credential.
     """
 
-    info = ctx.obj["info"]
     session = _get_session(ctx)
     serial = None
 
+    if upload:
+        raise CliFail(
+            "Automated YubiCloud upload support has been ended. "
+            "You can manually upload a credential by saving it as a CSV file "
+            "(use -O/--config-output) and then submitting it to "
+            "https://upload.yubico.com."
+        )
+
     if public_id and serial_public_id:
         ctx.fail("Invalid options: --public-id conflicts with --serial-public-id.")
 
     if private_id and generate_private_id:
         ctx.fail("Invalid options: --private-id conflicts with --generate-public-id.")
 
-    if upload and force:
-        ctx.fail("Invalid options: --upload conflicts with --force.")
-
     if key and generate_key:
         ctx.fail("Invalid options: --key conflicts with --generate-key.")
 
@@ -416,18 +431,20 @@ def yubiotp(
             try:
                 serial = session.get_serial()
             except CommandError:
-                raise CliFail("Serial number not set, public ID must be provided")
+                raise CliFail("Serial number not set, public ID must be provided.")
 
             public_id = modhex_encode(b"\xff\x00" + struct.pack(b">I", serial))
             click.echo(f"Using YubiKey serial as public ID: {public_id}")
         elif force:
             ctx.fail(
-                "Public ID not given. Please remove the --force flag, or "
+                "Public ID not given. Remove the --force flag, or "
                 "add the --serial-public-id flag or --public-id option."
             )
         else:
             public_id = click_prompt("Enter public ID")
 
+    if len(public_id) % 2:
+        ctx.fail("Invalid public ID, length must be a multiple of 2.")
     try:
         public_id = modhex_decode(public_id)
     except ValueError:
@@ -439,7 +456,7 @@ def yubiotp(
             click.echo(f"Using a randomly generated private ID: {private_id.hex()}")
         elif force:
             ctx.fail(
-                "Private ID not given. Please remove the --force flag, or "
+                "Private ID not given. Remove the --force flag, or "
                 "add the --generate-private-id flag or --private-id option."
             )
         else:
@@ -452,30 +469,13 @@ def yubiotp(
             click.echo(f"Using a randomly generated secret key: {key.hex()}")
         elif force:
             ctx.fail(
-                "Secret key not given. Please remove the --force flag, or "
+                "Secret key not given. Remove the --force flag, or "
                 "add the --generate-key flag or --key option."
             )
         else:
             key = click_prompt("Enter secret key")
             key = bytes.fromhex(key)
 
-    if upload:
-        click.confirm("Upload credential to YubiCloud?", abort=True, err=True)
-
-        try:
-            upload_url = _prepare_upload_key(
-                key,
-                public_id,
-                private_id,
-                serial=info.serial,
-                user_agent="ykman/" + __version__,
-            )
-            click.echo("Upload to YubiCloud initiated successfully.")
-            logger.info("Initiated YubiCloud upload")
-        except _PrepareUploadFailed as e:
-            error_msg = "\n".join(e.messages())
-            raise CliFail("Upload to YubiCloud failed.\n" + error_msg)
-
     force or click.confirm(
         f"Program a YubiOTP credential in slot {slot}?", abort=True, err=True
     )
@@ -497,12 +497,11 @@ def yubiotp(
         serial = serial or session.get_serial()
         csv = format_csv(serial, public_id, private_id, key, access_code)
         config_output.write(csv + "\n")
-        logger.info(f"Configuration parameters written to {_fname(config_output)}")
-
-    if upload:
-        logger.info("Launching browser for YubiCloud upload")
-        click.echo("Opening upload form in browser: " + upload_url)
-        webbrowser.open_new_tab(upload_url)
+        log_or_echo(
+            f"Configuration parameters written to {_fname(config_output)}",
+            logger,
+            config_output,
+        )
 
 
 @otp.command()
@@ -567,6 +566,7 @@ def static(ctx, slot, password, generate, length, keyboard_layout, no_enter, for
             ctx.obj["access_code"],
             ctx.obj["access_code"],
         )
+        click.echo(f"Static password stored in slot {slot}.")
     except CommandError:
         raise CliFail(_WRITE_FAIL_MSG)
 
@@ -617,7 +617,7 @@ def chalresp(ctx, slot, key, totp, touch, force, generate):
     else:
         if force and not generate:
             ctx.fail(
-                "No secret key given. Please remove the --force flag, "
+                "No secret key given. Remove the --force flag, "
                 "set the KEY argument or set the --generate flag."
             )
         elif generate:
@@ -652,6 +652,7 @@ def chalresp(ctx, slot, key, totp, touch, force, generate):
             ctx.obj["access_code"],
             ctx.obj["access_code"],
         )
+        click.echo(f"{cred_type} credential stored in slot {slot}.")
     except CommandError:
         raise CliFail(_WRITE_FAIL_MSG)
 
@@ -822,6 +823,7 @@ def hotp(ctx, slot, key, digits, counter, identifier, no_enter, force):
             ctx.obj["access_code"],
             ctx.obj["access_code"],
         )
+        click.echo(f"HOTP credential stored in slot {slot}.")
     except CommandError:
         raise CliFail(_WRITE_FAIL_MSG)
 
@@ -904,6 +906,8 @@ def settings(
             new_access_code = parse_access_code_hex(new_access_code)
         except Exception as e:
             ctx.fail("Failed to parse access code: " + str(e))
+        if ctx.obj["info"].pin_complexity and len(set(new_access_code)) < 2:
+            raise CliFail("Access code does not meet complexity requirement.")
 
     force or click.confirm(
         f"Update the settings for slot {slot}? "
@@ -911,7 +915,6 @@ def settings(
         abort=True,
         err=True,
     )
-    click.echo(f"Updating settings for slot {slot}...")
 
     pacing_bits = int(pacing or "0") // 20
     pacing_10ms = bool(pacing_bits & 1)
@@ -927,5 +930,6 @@ def settings(
             new_access_code,
             ctx.obj["access_code"],
         )
+        click.echo(f"Settings for slot {slot} updated.")
     except CommandError:
         raise CliFail(_WRITE_FAIL_MSG)
